#include "NvCaffeParser.h"
#include "NvInfer.h"

#include <cuda_runtime_api.h>
#include <iostream>
#include <vector>
#include <string>
#include <memory>
#include <assert.h>

class TensorRTLogger : public nvinfer1::ILogger
{
public:

    TensorRTLogger() : TensorRTLogger(Severity::kWARNING) {}

    TensorRTLogger(Severity severity) : reportableSeverity(severity) {}

    void log(Severity severity, const char* msg)
    {
        // suppress messages with severity enum value greater than the reportable
        if(severity > reportableSeverity) return;

        switch(severity)
        {
        case Severity::kINTERNAL_ERROR:
            std::cout << msg << std::endl;
            break;
        case Severity::kERROR:
            std::cout << msg << std::endl;
            break;
        case Severity::kWARNING:
            std::cout << msg << std::endl;
            break;
        case Severity::kINFO:
            std::cout << msg << std::endl;
            break;
        default:
            std::cout << msg << std::endl;
            break;
        }
    }

    Severity reportableSeverity;
};  // TensorRTLogger

class BufferManager
{
public:

    BufferManager(std::shared_ptr<nvinfer1::ICudaEngine> engine)
    {
        for(int i = 0; i < engine->getNbBindings(); i++)
        {
            auto dims = engine->getBindingDimensions(i);

            size_t vol = static_cast<size_t>(1) * sizeof(float);

            for(size_t i=0; i<dims.nbDims; ++i)
            {
                vol *= dims.d[i];
            }

            void *buffer = nullptr;

            cudaMalloc(&buffer, vol);

            mBuffers.push_back(buffer);
            mDeviceBindings.push_back(buffer);
        }
    }

    ~BufferManager()
    {
        for(size_t i=0; i<mBuffers.size(); ++i)
        {
            cudaFree(mBuffers[i]);
        }
    }

    std::vector<void *>& getDeviceBindings() { return mDeviceBindings; }

    std::vector<void *> mBuffers;
    std::vector<void *> mDeviceBindings;
};

struct InferDeleter
{
    template <typename T>
    void operator()(T* obj) const
    {
        if(obj)
        {
            obj->destroy();
        }
    }
};

template <typename T>
using SampleUniquePtr = std::unique_ptr<T, InferDeleter>;

int main(int argc, char** argv)
{
    const std::string filePrototxt   = "E:\\Git\\Model\\Caffe\\Caffe_MobileNet\\mobilenet_1.0_224_81.prototxt";
    const std::string fileCaffemodel = "E:\\Git\\Model\\Caffe\\Caffe_MobileNet\\mobilenet_1.0_224_81.caffemodel";
    std::vector<std::string> outputTensorNames ={ "relu1" };

    //const std::string filePrototxt   = "E:\\Lib\\nvidia\\TensorRT\\TensorRT-6.0.1.5\\data\\mnist\\mnist.prototxt";
    //const std::string fileCaffemodel = "E:\\Lib\\nvidia\\TensorRT\\TensorRT-6.0.1.5\\data\\mnist\\mnist.caffemodel";
    //std::vector<std::string> outputTensorNames ={ "prob" };

    std::shared_ptr<nvinfer1::ICudaEngine> engine{ nullptr };

    {
        TensorRTLogger logger;

        auto builder = SampleUniquePtr<nvinfer1::IBuilder>(nvinfer1::createInferBuilder(logger));
        if(!builder)
        {
            return false;
        }

        auto network = SampleUniquePtr<nvinfer1::INetworkDefinition>(builder->createNetwork());
        if(!network)
        {
            return false;
        }

        auto config = SampleUniquePtr<nvinfer1::IBuilderConfig>(builder->createBuilderConfig());
        if(!config)
        {
            return false;
        }

        auto parser = SampleUniquePtr<nvcaffeparser1::ICaffeParser>(nvcaffeparser1::createCaffeParser());
        if(!parser)
        {
            return false;
        }

        const nvcaffeparser1::IBlobNameToTensor* blobNameToTensor
            = parser->parse(
                filePrototxt.c_str(),
                fileCaffemodel.c_str(),
                *network,
                nvinfer1::DataType::kFLOAT
            );

        for(auto& s : outputTensorNames)
        {
            network->markOutput(*blobNameToTensor->find(s.c_str()));
        }

        builder->setMaxBatchSize(1);
        config->setMaxWorkspaceSize(16 << 20);

        engine = std::shared_ptr<nvinfer1::ICudaEngine>(
            builder->buildEngineWithConfig(*network, *config),
            InferDeleter()
            );

        if(!engine)
            return false;
    }

    SampleUniquePtr<nvinfer1::IExecutionContext> context;

    BufferManager buffers(engine);

    while(1)
    {
        std::cout << "inference"<<std::endl;

        context.reset(engine->createExecutionContext());

        if(!context)
        {
            return false;
        }

        if(!context->execute(1, buffers.getDeviceBindings().data()))
        {
            return false;
        }

        context.reset(nullptr);
    }

    return 0;
}
